@@ -5,10 +5,10 @@
  *  David Corcoran <corcoran@linuxnet.com>
  * Copyright (C) 2003-2004
  *  Damien Sauveron <damien.sauveron@labri.fr>
- * Copyright (C) 2002-2009
+ * Copyright (C) 2002-2011
  *  Ludovic Rousseau <ludovic.rousseau@free.fr>
  *
- * $Id: ifdhandler.h 4544 2009-11-12 14:33:44Z rousseau $
+ * $Id: ifdhandler.h 6302 2012-05-18 11:04:40Z rousseau $
  */
 
 /**
@@ -19,7 +19,7 @@
 The routines specified hereafter will allow you to write an IFD handler
 for the PC/SC Lite resource manager. Please use the complement
 developer's kit complete with headers and Makefile at:
-http://www.musclecard.com/drivers.html.
+http://www.musclecard.com/drivers.html
 
 This gives a common API for communication to most readers in a
 homogeneous fashion. This document assumes that the driver developer is
@@ -202,18 +202,23 @@ the device name of this port. It is dependent of the OS kernel. For
 example the first serial port device is called @c /dev/ttyS0 under Linux
 and @c /dev/cuaa0 under FreeBSD.
 
+If you want to use IFDHCreateChannel() instead of
+IFDHCreateChannelByName() then do not use any DEVICENAME line in the
+configuration file.  IFDHCreateChannel() will then be called with the
+CHANNELID parameter.
+
 @subsection LIBPATH
 The LIBPATH field is the filename of the driver code. The driver is a
 dynamically loaded piece of code (generally a @c drivername.so* file).
 
 @subsection CHANNELID
 The CHANNELID is no more used for recent drivers (IFD handler 3.0) and
-has been superseded by DEVICENAME. If you have an old driver this field
-is used to indicate the port to use. You should read your driver
-documentation to know what information is needed here. It should be the
-serial port number for a serial reader.
+has been superseded by DEVICENAME.
+
+If you have an old driver this field is used to indicate the port to
+use. You should read your driver documentation to know what information
+is needed here. It should be the serial port number for a serial reader.
 
-@subsection CHANNELID
 CHANNELID was the numeric version of the port in which the reader will
 be located. This may be done by a symbolic link where @c /dev/pcsc/1 is
 the first device which may be a symbolic link to @c /dev/ttyS0 or
@@ -226,11 +231,6 @@ whichever location your reader resides.
 
 #include <pcsclite.h>
 
-#ifdef __cplusplus
-extern "C"
-{
-#endif
-
 	/*
 	 * List of data structures available to ifdhandler
 	 */
@@ -302,8 +302,10 @@ extern "C"
 #define TAG_IFD_THREAD_SAFE             0x0FAD	/**< driver is thread safe */
 #define TAG_IFD_SLOTS_NUMBER            0x0FAE	/**< number of slots of the reader */
 #define TAG_IFD_SIMULTANEOUS_ACCESS     0x0FAF	/**< number of reader the driver can manage */
-#define TAG_IFD_POLLING_THREAD          0x0FB0	/**< driver uses a polling thread */
+#define TAG_IFD_POLLING_THREAD          0x0FB0	/**< not used. See TAG_IFD_POLLING_THREAD_WITH_TIMEOUT */
 #define TAG_IFD_POLLING_THREAD_KILLABLE 0x0FB1	/**< the polling thread can be killed */
+#define TAG_IFD_STOP_POLLING_THREAD     0x0FB2	/**< method used to stop the polling thread (instead of just pthread_kill()) */
+#define TAG_IFD_POLLING_THREAD_WITH_TIMEOUT 0x0FB3	/**< driver uses a polling thread with a timeout parameter */
 
 	/*
 	 * IFD Handler version number enummerations
@@ -424,25 +426,29 @@ port used by each reader.
   So it is something like: <tt>usb:08e6/3437:libusb:001:042</tt> under
   GNU/Linux.
 
-- libhal
+- libudev
 
-  If pcscd is compiled with libhal support instead of libusb (default
-  since pcsc-lite 1.4.100) the string will look like:
+  If pcscd is compiled with libudev support instead of libusb (default
+  since pcsc-lite 1.6.8) the string will look like:
 
   @code
-  printf("usb:%04x/%04x:libhal:%s", idVendor, idProduct, udi)
+  printf("usb:%04x/%04x:libudev:%d:%s", idVendor, idProduct,
+		bInterfaceNumber, devpath);
   @endcode
 
-  udi is the Universal Device Id at the HAL level.
+  bInterfaceNumber is the number of the interface on the device. It is
+  only usefull for devices with more than one CCID interface.
+
+  devpath is the filename of the device on the file system.
 
   So it is something like:
-  <tt>usb:08e6/3437:libhal:/org/freedesktop/Hal/devices/usb_device_8e6_3437_noserial_if0</tt>
+  <tt>usb:08e6/3437:libudev:0:/dev/bus/usb/008/047</tt>
   under GNU/Linux.
 
 - other
 
   If the driver does not understand the <tt>:libusb:</tt> or
-  <tt>:libhal:</tt> scheme or if a new scheme is used, the driver should
+  <tt>:libudev:</tt> scheme or if a new scheme is used, the driver should
   ignore the part it does not understand instead of failing.
 
   The driver shall recognize the <tt>usb:VID/PID</tt> part and, only if
@@ -577,6 +583,8 @@ don't mind loading a new driver for each reader then ignore Lun.
 @param[in] Tag Tag of the desired data value
 - \ref TAG_IFD_ATR
   Return the ATR and its size (implementation is mandatory).
+- \ref TAG_IFD_SLOTNUM
+  Unused/deprecated
 - \ref SCARD_ATTR_ATR_STRING
   Same as \ref TAG_IFD_ATR but this one is not mandatory. It is defined
   in Microsoft PC/SC SCardGetAttrib().
@@ -598,9 +606,22 @@ don't mind loading a new driver for each reader then ignore Lun.
   <tt>Value[0] = 1</tt> indicates the driver supports simultaneous slot
   accesses.
 - \ref TAG_IFD_POLLING_THREAD
-  If the driver provide a polling thread then @p Value is a pointer to this function
+  Unused/deprecated
+- \ref TAG_IFD_POLLING_THREAD_WITH_TIMEOUT
+  If the driver provides a polling thread then @p Value is a pointer to
+  this function. The function prototype is:
+@verbatim
+  RESPONSECODE foo(DWORD Lun, int timeout);
+@endverbatim
 - \ref TAG_IFD_POLLING_THREAD_KILLABLE
-  Tell if the polling thread can be killed (SYS_ThreadCancel()) by pcscd
+  Tell if the polling thread can be killed (pthread_kill()) by pcscd
+- \ref TAG_IFD_STOP_POLLING_THREAD
+  Returns a pointer in @p Value to the function used to stop the polling
+  thread returned by \ref TAG_IFD_POLLING_THREAD_WITH_TIMEOUT. The
+  function prototype is:
+@verbatim
+  RESPONSECODE foo(DWORD Lun);
+@endverbatim
 @param[in,out] Length Length of the desired data value
 @param[out] Value Value of the desired data
 
@@ -620,10 +641,6 @@ don't mind loading a new driver for each reader then ignore @p Lun.
 @ingroup IFDHandler
 @param[in] Lun Logical Unit Number
 @param[in] Tag Tag of the desired data value
-- \ref TAG_IFD_SLOTNUM
-  This is used in IFDHandler v1.0 to select the slot to use for the next
-  IFD_* command. This tag is no more used with versions 2.0 and 3.0 of
-  the IFD Handler.
 @param[in,out] Length Length of the desired data value
 @param[out] Value Value of the desired data
 
@@ -778,8 +795,4 @@ command each time this function is called.
  */
 RESPONSECODE IFDHICCPresence(DWORD Lun);
 
-#ifdef __cplusplus
-}
-#endif
-
 #endif
