@@ -1,6 +1,6 @@
 /*
     scardcontrol.c: sample code to use/test SCardControl() API
-    Copyright (C) 2004-2009   Ludovic Rousseau
+    Copyright (C) 2004-2011   Ludovic Rousseau
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
 */
 
 /*
- * $Id: scardcontrol.c 4927 2010-05-07 12:38:40Z rousseau $
+ * $Id: scardcontrol.c 6075 2011-10-24 12:28:16Z rousseau $
  */
 
 #include <stdio.h>
@@ -138,6 +138,13 @@ static void parse_properties(unsigned char *bRecvBuffer, int length)
 					putchar(p[i]);
 				printf(NORMAL "\n");
 				break;
+			case PCSCv2_PART10_PROPERTY_bPPDUSupport:
+				PRINT_GREEN_HEX2(" bPPDUSupport", value);
+				if (value & 1)
+					printf("  PPDU is supported over SCardControl using FEATURE_CCID_ESC_COMMAND\n");
+				if (value & 2)
+					printf("  PPDU is supported over SCardTransmit\n");
+				break;
 			default:
 				printf(" Unknown tag: 0x%02X (length = %d)\n", tag, len);
 		}
@@ -211,8 +218,9 @@ int main(int argc, char *argv[])
 	DWORD pin_properties_ioctl = 0;
 	DWORD mct_readerdirect_ioctl = 0;
 	DWORD properties_in_tlv_ioctl = 0;
+	DWORD ccid_esc_command = 0;
 	SCARD_IO_REQUEST pioRecvPci;
- 	SCARD_IO_REQUEST pioSendPci;
+	SCARD_IO_REQUEST pioSendPci;
 	PCSC_TLV_STRUCTURE *pcsc_tlv;
 #if defined(VERIFY_PIN) | defined(MODIFY_PIN)
 	int offset;
@@ -223,7 +231,6 @@ int main(int argc, char *argv[])
 #ifdef MODIFY_PIN
 	PIN_MODIFY_STRUCTURE *pin_modify;
 #endif
-	char secoder_info[] = { 0x20, 0x70, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00 };
 
 	printf("SCardControl sample code\n");
 	printf("V 1.4 © 2004-2010, Ludovic Rousseau <ludovic.rousseau@free.fr>\n\n");
@@ -305,7 +312,7 @@ int main(int argc, char *argv[])
 	/* connect to a reader (even without a card) */
 	dwActiveProtocol = -1;
 	printf("Using reader: " GREEN "%s\n" NORMAL, readers[reader_nb]);
-	rv = SCardConnect(hContext, readers[reader_nb], SCARD_SHARE_DIRECT,
+	rv = SCardConnect(hContext, readers[reader_nb], SCARD_SHARE_SHARED,
 		SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, &hCard, &dwActiveProtocol);
 	printf(" Protocol: " GREEN "%ld\n" NORMAL, dwActiveProtocol);
 	PCSC_ERROR_EXIT(rv, "SCardConnect")
@@ -376,6 +383,10 @@ int main(int argc, char *argv[])
 				PRINT_GREEN("Reader supports", "FEATURE_GET_TLV_PROPERTIES");
 				properties_in_tlv_ioctl = ntohl(pcsc_tlv[i].value);
 				break;
+			case FEATURE_CCID_ESC_COMMAND:
+				PRINT_GREEN("Reader supports", "FEATURE_CCID_ESC_COMMAND");
+				ccid_esc_command = ntohl(pcsc_tlv[i].value);
+				break;
 			default:
 				printf("Can't parse tag: " RED "0x%02X" NORMAL, pcsc_tlv[i].tag);
 		}
@@ -386,8 +397,8 @@ int main(int argc, char *argv[])
 	{
 		int value;
 
-		rv = SCardControl(hCard, properties_in_tlv_ioctl, secoder_info,
-			sizeof(secoder_info), bRecvBuffer, sizeof(bRecvBuffer), &length);
+		rv = SCardControl(hCard, properties_in_tlv_ioctl, NULL, 0,
+			bRecvBuffer, sizeof(bRecvBuffer), &length);
 		PCSC_ERROR_CONT(rv, "SCardControl(GET_TLV_PROPERTIES)")
 
 		printf("GET_TLV_PROPERTIES (" GREEN "%ld" NORMAL "): " GREEN, length);
@@ -410,6 +421,8 @@ int main(int argc, char *argv[])
 
 	if (mct_readerdirect_ioctl)
 	{
+		char secoder_info[] = { 0x20, 0x70, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00 };
+
 		rv = SCardControl(hCard, mct_readerdirect_ioctl, secoder_info,
 			sizeof(secoder_info), bRecvBuffer, sizeof(bRecvBuffer), &length);
 		PCSC_ERROR_CONT(rv, "SCardControl(MCT_READER_DIRECT)")
@@ -420,6 +433,50 @@ int main(int argc, char *argv[])
 		printf("\n");
 	}
 
+	if (pin_properties_ioctl)
+	{
+		PIN_PROPERTIES_STRUCTURE *pin_properties;
+
+		rv = SCardControl(hCard, pin_properties_ioctl, NULL, 0,
+			bRecvBuffer, sizeof(bRecvBuffer), &length);
+		PCSC_ERROR_CONT(rv, "SCardControl(pin_properties_ioctl)")
+
+		printf("PIN PROPERTIES (" GREEN "%ld" NORMAL "): " GREEN, length);
+		for (i=0; i<length; i++)
+			printf("%02X ", bRecvBuffer[i]);
+		printf(NORMAL "\n");
+
+		pin_properties = (PIN_PROPERTIES_STRUCTURE *)bRecvBuffer;
+		PRINT_GREEN_HEX4(" wLcdLayout", pin_properties -> wLcdLayout);
+		PRINT_GREEN_DEC(" bEntryValidationCondition", pin_properties ->	bEntryValidationCondition);
+		PRINT_GREEN_DEC(" bTimeOut2", pin_properties -> bTimeOut2);
+
+		printf("\n");
+	}
+
+#ifdef GET_GEMPC_FIRMWARE
+	if (ccid_esc_command)
+	{
+		/* get GemPC firmware */
+		printf("Get GemPC Firmware\n");
+
+		/* this is specific to Gemalto readers */
+		bSendBuffer[0] = 0x02;
+		rv = SCardControl(hCard, ccid_esc_command, bSendBuffer,
+			1, bRecvBuffer, sizeof(bRecvBuffer), &length);
+
+		printf(" Firmware: " GREEN);
+		for (i=0; i<length; i++)
+			printf("%02X ", bRecvBuffer[i]);
+		printf(NORMAL "\n");
+
+		bRecvBuffer[length] = '\0';
+		printf(" Firmware: " GREEN "%s" NORMAL" (length " GREEN "%ld" NORMAL " bytes)\n", bRecvBuffer, length);
+
+		PCSC_ERROR_CONT(rv, "SCardControl")
+	}
+#endif
+
 	if (0 == verify_ioctl)
 	{
 		printf("Reader %s does not support PIN verification\n",
@@ -495,16 +552,21 @@ int main(int argc, char *argv[])
 	printf(" Secure verify PIN\n");
 	pin_verify = (PIN_VERIFY_STRUCTURE *)bSendBuffer;
 
+	/* table for bEntryValidationCondition
+	 * 0x01: Max size reached
+	 * 0x02: Validation key pressed
+	 * 0x04: Timeout occured
+	 */
 	/* PC/SC v2.02.05 Part 10 PIN verification data structure */
 	pin_verify -> bTimerOut = 0x00;
 	pin_verify -> bTimerOut2 = 0x00;
 	pin_verify -> bmFormatString = 0x82;
 	pin_verify -> bmPINBlockString = 0x04;
 	pin_verify -> bmPINLengthFormat = 0x00;
-	pin_verify -> wPINMaxExtraDigit = HOST_TO_CCID_16(0x0408); /* Min Max */
+	pin_verify -> wPINMaxExtraDigit = 0x0408; /* Min Max */
 	pin_verify -> bEntryValidationCondition = 0x02;	/* validation key pressed */
 	pin_verify -> bNumberMessage = 0x01;
-	pin_verify -> wLangId = HOST_TO_CCID_16(0x0904);
+	pin_verify -> wLangId = 0x0904;
 	pin_verify -> bMsgIndex = 0x00;
 	pin_verify -> bTeoPrologue[0] = 0x00;
 	pin_verify -> bTeoPrologue[1] = 0x00;
@@ -526,7 +588,7 @@ int main(int argc, char *argv[])
 	pin_verify -> abData[offset++] = 0x00;	/* '\0' */
 	pin_verify -> abData[offset++] = 0x00;	/* '\0' */
 	pin_verify -> abData[offset++] = 0x00;	/* '\0' */
-	pin_verify -> ulDataLength = HOST_TO_CCID_32(offset);	/* APDU size */
+	pin_verify -> ulDataLength = offset;	/* APDU size */
 
 	length = sizeof(PIN_VERIFY_STRUCTURE) + offset -1;	/* -1 because PIN_VERIFY_STRUCTURE contains the first byte of abData[] */
 
@@ -628,23 +690,28 @@ int main(int argc, char *argv[])
 	 * bConfirmPIN = 1, bNumberMessage = 2: "New Pin" "Confirm Pin"
 	 * bConfirmPIN = 0, bNumberMessage = 1: "New Pin"
 	 */
+	/* table for bMsgIndex[1-3]
+	 * 00: PIN insertion prompt        “ENTER SMARTCARD PIN”
+	 * 01: PIN Modification prompt     “ ENTER NEW PIN”
+	 * 02: NEW PIN Confirmation prompt “ CONFIRM NEW PIN”
+	 */
 	/* PC/SC v2.02.05 Part 10 PIN modification data structure */
 	pin_modify -> bTimerOut = 0x00;
 	pin_modify -> bTimerOut2 = 0x00;
 	pin_modify -> bmFormatString = 0x82;
 	pin_modify -> bmPINBlockString = 0x04;
 	pin_modify -> bmPINLengthFormat = 0x00;
-	pin_modify -> bInsertionOffsetOld = 0x00; 	/* offset from APDU start */
+	pin_modify -> bInsertionOffsetOld = 0x00;	/* offset from APDU start */
 	pin_modify -> bInsertionOffsetNew = 0x04;	/* offset from APDU start */
-	pin_modify -> wPINMaxExtraDigit = HOST_TO_CCID_16(0x0408);	/* Min Max */
+	pin_modify -> wPINMaxExtraDigit = 0x0408;	/* Min Max */
 	pin_modify -> bConfirmPIN = 0x03;	/* b0 set = confirmation requested */
 									/* b1 set = current PIN entry requested */
 	pin_modify -> bEntryValidationCondition = 0x02;	/* validation key pressed */
 	pin_modify -> bNumberMessage = 0x03; /* see table above */
-	pin_modify -> wLangId = HOST_TO_CCID_16(0x0904);
+	pin_modify -> wLangId = 0x0904;
 	pin_modify -> bMsgIndex1 = 0x00;
-	pin_modify -> bMsgIndex2 = 0x00;
-	pin_modify -> bMsgIndex3 = 0x00;
+	pin_modify -> bMsgIndex2 = 0x01;
+	pin_modify -> bMsgIndex3 = 0x02;
 	pin_modify -> bTeoPrologue[0] = 0x00;
 	pin_modify -> bTeoPrologue[1] = 0x00;
 	pin_modify -> bTeoPrologue[2] = 0x00;
@@ -665,7 +732,7 @@ int main(int argc, char *argv[])
 	pin_modify -> abData[offset++] = 0x30;	/* '0' */
 	pin_modify -> abData[offset++] = 0x30;	/* '0' */
 	pin_modify -> abData[offset++] = 0x30;	/* '0' */
-	pin_modify -> ulDataLength = HOST_TO_CCID_32(offset);	/* APDU size */
+	pin_modify -> ulDataLength = offset;	/* APDU size */
 
 	length = sizeof(PIN_MODIFY_STRUCTURE) + offset -1;	/* -1 because PIN_MODIFY_STRUCTURE contains the first byte of abData[] */
 
